import path from 'path';
import ts from 'typescript';
import { extractDefaultComponent, resolveComponent, extractCsf3Stories } from '@previewjs/csf3';
import { EMPTY_OBJECT_TYPE, UNKNOWN_TYPE, dereferenceType, objectType, maybeOptionalType, helpers, NODE_TYPE, functionType, arrayType } from '@previewjs/type-analyzer';

function analyzePreactComponent(typeResolver, signature) {
  const firstParam = signature.getParameters()[0];
  if (!firstParam) {
    return {
      propsType: EMPTY_OBJECT_TYPE,
      types: {}
    };
  }
  if (!firstParam.valueDeclaration) {
    return {
      propsType: UNKNOWN_TYPE,
      types: {}
    };
  }
  const type = typeResolver.checker.getTypeOfSymbolAtLocation(
    firstParam,
    firstParam.valueDeclaration
  );
  try {
    let { type: propsType, collected: types } = typeResolver.resolveType(type);
    [propsType] = dereferenceType(propsType, types, []);
    stripUnusedProps:
      if (propsType.kind === "object" && ts.isParameter(firstParam.valueDeclaration)) {
        if (ts.isObjectBindingPattern(firstParam.valueDeclaration.name)) {
          const bindingPattern = firstParam.valueDeclaration.name;
          const usedProps = /* @__PURE__ */ new Set();
          const propsWithDefault = /* @__PURE__ */ new Set();
          for (const element of bindingPattern.elements) {
            if (element.dotDotDotToken) {
              break stripUnusedProps;
            }
            const elementName = element.propertyName || element.name;
            if (!ts.isIdentifier(elementName)) {
              break stripUnusedProps;
            }
            usedProps.add(elementName.text);
            if (element.initializer) {
              propsWithDefault.add(elementName.text);
            }
          }
          propsType = objectType(
            Object.fromEntries(
              Object.entries(propsType.fields).filter(([key]) => usedProps.has(key)).map(([key, type2]) => [
                key,
                maybeOptionalType(type2, propsWithDefault.has(key))
              ])
            )
          );
        }
      }
    if (propsType.kind === "object") {
      propsType = {
        kind: "object",
        fields: Object.fromEntries(
          Object.entries(propsType.fields).filter(
            ([key]) => !["jsx", "key", "ref"].includes(key)
          )
        )
      };
    }
    return { propsType, types };
  } catch (e) {
    console.warn(
      `Unable to resolve props type for ${typeResolver.checker.typeToString(
        type
      )}`,
      e
    );
  }
  return {
    propsType: UNKNOWN_TYPE,
    types: {}
  };
}

function extractPreactComponents(resolver, absoluteFilePath) {
  const sourceFile = resolver.sourceFile(absoluteFilePath);
  if (!sourceFile) {
    return [];
  }
  const functions = [];
  for (const statement of sourceFile.statements) {
    if (ts.isExportAssignment(statement)) {
      if (ts.isIdentifier(statement.expression)) {
        continue;
      }
      functions.push(["default", statement, statement.expression]);
    } else if (ts.isVariableStatement(statement)) {
      for (const declaration of statement.declarationList.declarations) {
        if (!ts.isIdentifier(declaration.name) || !declaration.initializer) {
          continue;
        }
        functions.push([
          declaration.name.text,
          statement,
          declaration.initializer
        ]);
      }
    } else if (ts.isFunctionDeclaration(statement)) {
      const isDefaultExport = !!statement.modifiers?.find(
        (m) => m.kind === ts.SyntaxKind.ExportKeyword
      ) && !!statement.modifiers?.find(
        (m) => m.kind === ts.SyntaxKind.DefaultKeyword
      );
      const name = statement.name?.text;
      if (isDefaultExport || name) {
        functions.push([name || "default", statement, statement]);
      }
    } else if (ts.isClassDeclaration(statement) && statement.name) {
      functions.push([statement.name.text, statement, statement]);
    }
  }
  const storiesDefaultComponent = extractDefaultComponent(sourceFile);
  const resolvedStoriesComponent = storiesDefaultComponent ? resolveComponent(resolver.checker, storiesDefaultComponent) : null;
  const components = [];
  const args = helpers.extractArgs(sourceFile);
  const nameToExportedName = helpers.extractExportedNames(sourceFile);
  function extractComponentTypeInfo(node, name) {
    const hasArgs = !!args[name];
    const isExported = name === "default" || !!nameToExportedName[name];
    if (storiesDefaultComponent && hasArgs && isExported) {
      return { kind: "story", associatedComponent: resolvedStoriesComponent };
    }
    const signature = extractComponentSignature(resolver.checker, node);
    if (signature) {
      return {
        kind: "component",
        exported: isExported,
        analyze: async () => analyzePreactComponent(resolver, signature)
      };
    }
    return null;
  }
  for (const [name, statement, node] of functions) {
    const info = extractComponentTypeInfo(node, name);
    if (info) {
      components.push({
        absoluteFilePath,
        name,
        offsets: [[statement.getStart(), statement.getEnd()]],
        info
      });
    }
  }
  return [...components, ...extractCsf3Stories(resolver, sourceFile)];
}
function extractComponentSignature(checker, node) {
  const type = checker.getTypeAtLocation(node);
  for (const callSignature of type.getCallSignatures()) {
    if (isValidComponentReturnType(callSignature.getReturnType())) {
      return callSignature;
    }
  }
  if (type.symbol) {
    const classType = checker.getTypeOfSymbolAtLocation(type.symbol, node);
    for (const constructSignature of classType.getConstructSignatures()) {
      const returnType = constructSignature.getReturnType();
      if (returnType.getProperty("render")) {
        return constructSignature;
      }
    }
  }
  return null;
}
function isValidComponentReturnType(type) {
  if (isJsxElement(type)) {
    return true;
  }
  return false;
}
const jsxElementTypes = /* @__PURE__ */ new Set(["Element", "VNode"]);
function isJsxElement(type) {
  if (type.isUnion()) {
    for (const subtype of type.types) {
      if (isJsxElement(subtype)) {
        return true;
      }
    }
  }
  return jsxElementTypes.has(type.symbol?.getEscapedName().toString());
}

const PREACT_SPECIAL_TYPES = {
  Component: NODE_TYPE,
  ComponentClass: functionType(NODE_TYPE),
  ComponentChildren: NODE_TYPE,
  FunctionComponent: functionType(NODE_TYPE),
  FunctionalComponent: functionType(NODE_TYPE),
  PreactElement: NODE_TYPE,
  PreactNode: NODE_TYPE,
  PreactNodeArray: arrayType(NODE_TYPE),
  PreactPortal: NODE_TYPE
};

const preactFrameworkPlugin = {
  isCompatible: async (dependencies) => {
    const version = await dependencies["preact"]?.readInstalledVersion();
    if (!version) {
      return false;
    }
    return parseInt(version) >= 10;
  },
  async create() {
    const previewDirPath = path.resolve(__dirname, "..", "preview");
    return {
      pluginApiVersion: 3,
      name: "@previewjs/plugin-preact",
      defaultWrapperPath: "__previewjs__/Wrapper.tsx",
      previewDirPath,
      specialTypes: PREACT_SPECIAL_TYPES,
      tsCompilerOptions: {
        jsx: ts.JsxEmit.ReactJSX,
        jsxImportSource: "preact",
        jsxFactory: "h",
        jsxFragmentFactory: "Fragment"
      },
      detectComponents: async (typeAnalyzer, absoluteFilePaths) => {
        const resolver = typeAnalyzer.analyze(absoluteFilePaths);
        const components = [];
        for (const absoluteFilePath of absoluteFilePaths) {
          components.push(
            ...extractPreactComponents(resolver, absoluteFilePath)
          );
        }
        return components;
      },
      viteConfig: () => {
        return {
          resolve: {
            alias: {
              react: "preact/compat",
              "react-dom/test-utils": "preact/test-utils",
              "react-dom": "preact/compat",
              "react/jsx-runtime": "preact/jsx-runtime"
            }
          },
          esbuild: {
            jsx: "automatic",
            jsxImportSource: "preact",
            jsxFactory: "h",
            jsxFragment: "fragment"
          },
          plugins: [
            {
              name: "previewjs:optimize-deps",
              config: () => ({
                optimizeDeps: {
                  include: ["preact", "preact/jsx-runtime"]
                }
              })
            }
          ]
        };
      },
      incompatibleVitePlugins: ["prefresh", "vite:preact-jsx"]
    };
  }
};

export { preactFrameworkPlugin as default };
